package dhcp

import (
	"bytes"
	"encoding/binary"
	"fmt"
	"io"
	"net/netip"
	"strings"

	"git.wntrmute.dev/kyle/goutils/assert"
	log "git.wntrmute.dev/kyle/goutils/log"
	"git.wntrmute.dev/kyle/kdhcp/leases"
)

const (
	maxHardwareAddrLen = 16
	maxServerName      = 64
	maxFileName        = 128
)

const (
	MessageTypeBootRequest  = 1
	MessageTypeBootResponse = 2
)

func formatMAC(mac []byte) string {
	s := []string{}
	for i := 0; i < len(mac); i++ {
		s = append(s, fmt.Sprintf("%0x", mac[i:i+1]))
	}

	return strings.Join(s, ":")
}

func readIPv4(r io.Reader) (netip.Addr, error) {
	var buf [4]byte

	n, err := r.Read(buf[:])
	if err != nil {
		return netip.Addr{}, fmt.Errorf("dhcp: while reading IPv4 address from reader: %w", err)
	}

	assert.Bool(n == 4, fmt.Sprintf("read %d bytes, but expected to read 4 bytes", n))
	return netip.AddrFrom4(buf), nil
}

func newPacketReaderFunc(r io.Reader) func(v any) error {
	return func(v any) error {
		return binary.Read(r, binary.BigEndian, v)
	}
}

func newPacketWriterFunc(w io.Writer) func(v any) error {
	return func(v any) error {
		return binary.Write(w, binary.BigEndian, v)
	}
}

type Packet struct {
	MessageType     uint8
	HardwareType    uint8
	HardwareAddress leases.HardwareAddress
	Hops            uint8
	TransactionID   uint32
	SecondsElapsed  uint16
	Flags           uint16
	ServerName      string
	FileName        string

	ClientIP netip.Addr
	YourIP   netip.Addr
	NextIP   netip.Addr
	RelayIP  netip.Addr

	DHCPType          DHCPMessageType // option 53
	HostName          string          // option 12
	ClientID          string          // option 61
	ParameterRequests []OptionTag
	Parameters        []Parameter
	endOptions        bool
}

func (pkt *Packet) Read(packet []byte) (err error) {
	buf := bytes.NewBuffer(packet)
	read := newPacketReaderFunc(buf)

	if err = read(&pkt.MessageType); err != nil {
		return err
	}

	if err = read(&pkt.HardwareType); err != nil {
		return err
	}

	var hwaLength uint8
	if err = read(&hwaLength); err != nil {
		return err
	}

	if err = read(&pkt.Hops); err != nil {
		return err
	}

	if err = read(&pkt.TransactionID); err != nil {
		return err
	}

	if err = read(&pkt.SecondsElapsed); err != nil {
		return err
	}

	if err = read(&pkt.Flags); err != nil {
		return err
	}

	if pkt.ClientIP, err = readIPv4(buf); err != nil {
		return err
	}

	if pkt.YourIP, err = readIPv4(buf); err != nil {
		return err
	}

	if pkt.NextIP, err = readIPv4(buf); err != nil {
		return err
	}

	if pkt.RelayIP, err = readIPv4(buf); err != nil {
		return err
	}

	pkt.HardwareAddress = make([]byte, int(hwaLength))
	if _, err = buf.Read(pkt.HardwareAddress); err != nil {
		return err
	}

	hwaPad := make([]byte, maxHardwareAddrLen-hwaLength)
	if _, err = buf.Read(hwaPad); err != nil {
		return err
	}

	tempBuf := make([]byte, maxServerName)
	if _, err = buf.Read(tempBuf); err != nil {
		return err
	}
	pkt.ServerName = string(bytes.Trim(tempBuf, "\x00"))

	tempBuf = make([]byte, maxFileName)
	if _, err = buf.Read(tempBuf); err != nil {
		return err
	}
	pkt.FileName = string(bytes.Trim(tempBuf, "\x00"))

	if err = ReadMagicCookie(buf); err != nil {
		return err
	}

	for {
		if pkt.endOptions {
			break
		}

		tag, err := buf.ReadByte()
		if err != nil {
			if err == io.EOF {
				break
			}
			return err
		}

		err = ReadOption(pkt, tag, buf)
		if err != nil {
			log.Spew(*pkt)
			log.Spew(packet)
			return err
		}
	}
	return nil
}

func ReadPacket(pkt []byte) (*Packet, error) {
	packet := &Packet{}
	err := packet.Read(pkt)
	if err != nil {
		return nil, err
	}

	log.Debugf("dhcp: BOOTP packet with txid %d for %s", packet.TransactionID, formatMAC(packet.HardwareAddress))
	return packet, nil
}

func (pkt *Packet) Write(w io.Writer) error {
	write := newPacketWriterFunc(w)

	if err := write(pkt.MessageType); err != nil {
		return err
	}

	if err := write(pkt.HardwareType); err != nil {
		return err
	}

	if err := write(uint8(len(pkt.HardwareAddress))); err != nil {
		return err
	}

	if err := write(pkt.Hops); err != nil {
		return err
	}

	if err := write(pkt.TransactionID); err != nil {
		return err
	}

	if err := write(pkt.SecondsElapsed); err != nil {
		return err
	}

	if err := write(pkt.Flags); err != nil {
		return err
	}

	if err := write(pkt.ClientIP.AsSlice()); err != nil {
		return err
	}

	if err := write(pkt.YourIP.AsSlice()); err != nil {
		return err
	}

	if err := write(pkt.NextIP.AsSlice()); err != nil {
		return err
	}

	if err := write(pkt.RelayIP.AsSlice()); err != nil {
		return err
	}

	padding := make([]byte, maxHardwareAddrLen-len(pkt.HardwareAddress))
	if err := write(pkt.HardwareAddress); err != nil {
		return err
	}

	if err := write(padding); err != nil {
		return err
	}

	padding = make([]byte, maxServerName-len(pkt.ServerName))
	if err := write([]byte(pkt.ServerName)); err != nil {
		return err
	}

	if err := write(padding); err != nil {
		return err
	}

	padding = make([]byte, maxFileName-len(pkt.FileName))
	if err := write([]byte(pkt.FileName)); err != nil {
		return err
	}

	if err := write(padding); err != nil {
		return err
	}

	if err := write(magicCookie); err != nil {
		return err
	}

	// TODO: write parameter request write

	for _, p := range pkt.Parameters {
		if !p.Valid() {
			continue
		}

		if err := write(p.Bytes()); err != nil {
			return err
		}
	}

	if err := write(OptionTagEnd); err != nil {
		return err
	}

	return nil
}

func WritePacket(pkt *Packet) ([]byte, error) {
	buf := &bytes.Buffer{}
	if err := pkt.Write(buf); err != nil {
		return nil, err
	}

	return buf.Bytes(), nil
}
