package leases

import (
	"fmt"
	"net/netip"
	"sort"
	"strings"
	"time"
)

type Info struct {
	HostName        string          `yaml:"hostname" json:"hostname,omitempty"`
	Addr            netip.Addr      `yaml:"addr" json:"addr,omitempty"`
	HardwareAddress HardwareAddress `yaml:"mac_addr" json:"hardware_address,omitempty"`
	Expires         time.Time       `yaml:"expires" json:"expires,omitempty"`
}

func (li *Info) String() string {
	return fmt.Sprintf("lease[hostname=%s addr=%s hw=%s expires=%s]", li.HostName, li.Addr, li.HardwareAddress, li.Expires)
}

type sortableLease []*Info

func (a sortableLease) Len() int           { return len(a) }
func (a sortableLease) Swap(i, j int)      { a[i], a[j] = a[j], a[i] }
func (a sortableLease) Less(i, j int) bool { return a[i].Addr.Less(a[j].Addr) }

func (li *Info) ResetExpiry(t time.Time, dur time.Duration) {
	li.Expires = t.Add(dur)
}

func (li *Info) IsExpired(t time.Time) bool {
	return t.After(li.Expires)
}

func (li *Info) Expire() {
	li.Expires = time.Time{}
}

func Sort(leases []*Info) []*Info {
	sortable := sortableLease(leases)
	sort.Sort(sortable)

	return []*Info(sortable)
}

func (lease *Info) Reset() *Info {
	lease.Expires = time.Time{}
	lease.HardwareAddress = nil
	return lease
}

func (lease Info) Compare(other Info) error {
	susFields := []string{}

	if lease.Addr != other.Addr {
		susFields = append(susFields, fmt.Sprintf("address is %s, but is recorded as %s", lease.Addr, other.Addr))
	}

	if !lease.HardwareAddress.Match(other.HardwareAddress) {
		susFields = append(susFields, fmt.Sprintf("hardware address is %s, but is recorded as %s", lease.HardwareAddress, other.HardwareAddress))
	}

	if lease.HostName != other.HostName {
		susFields = append(susFields, fmt.Sprintf("hostname is %s, but is recorded as %s", lease.HostName, other.HostName))
	}

	if len(susFields) > 0 {
		return fmt.Errorf("suspicious lease: %s", strings.Join(susFields, ";"))
	}

	return nil
}
